'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var solidJs = require('solid-js');
var deep = require('@solid-primitives/deep');
var isEqual = require('fast-deep-equal');
var web = require('solid-js/web');
var resizeObserver = require('@solid-primitives/resize-observer');

/**
 * Enhanced random function, selects a random value between a minimum and maximum. If the values provided are both
 * integers then the number returned will be an integer, otherwise the return number will be a decimal.
 * @param min The minimum value
 * @param max The maximum value
 */
function random(min, max) {
  const randomNumber = Math.random() * (max - min + 1) + min;
  if (!Number.isInteger(min) || !Number.isInteger(max)) {
    return randomNumber;
  } else {
    return Math.floor(randomNumber);
  }
}

/**
 * Linear interpolation function to gradually step towards a target value
 * @param start The current value
 * @param end The target value
 * @param normal The rate of change between 0 and 1 (0 = no change, 1 = instant)
 */
function lerp(start, end, normal) {
  return (1 - normal) * start + normal * end;
}

/**
 * Gets the height and width of the provided HTML element
 * @param element The html element to measure
 */
function getSize(element) {
  if (!element) return {
    height: 0,
    width: 0
  };
  return {
    height: element.offsetHeight,
    width: element.offsetWidth
  };
}

/**
 * Selects a random item from an array of inputs.
 *
 * @param items The array of items to choose from
 * @returns A random item selected from the array
 */
function randomElement(items) {
  const index = Math.floor(Math.random() * items.length);
  return items[index];
}

const defaultConfig = {
  color: "#dee4fd",
  radius: [0.5, 3.0],
  speed: [1.0, 3.0],
  wind: [-0.5, 2.0],
  changeFrequency: 200,
  rotationSpeed: [-1.0, 1.0]
};
/**
 * An individual snowflake that will update it's location every call to `update`
 * and draw itself to the canvas every call to `draw`.
 */
class Snowflake {
  static offscreenCanvases = new WeakMap();
  constructor(canvas, config = {}) {
    // Set custom config
    this.updateConfig(config);

    // Setting initial parameters
    const {
      radius,
      wind,
      speed,
      rotationSpeed
    } = this.config;
    this.params = {
      x: random(0, canvas.offsetWidth),
      y: random(-canvas.offsetHeight, 0),
      rotation: random(0, 360),
      radius: random(...radius),
      speed: random(...speed),
      wind: random(...wind),
      rotationSpeed: random(...rotationSpeed),
      nextSpeed: random(...wind),
      nextWind: random(...speed),
      nextRotationSpeed: random(...rotationSpeed)
    };
    this.framesSinceLastUpdate = 0;
  }
  selectImage() {
    if (this.config.images && this.config.images.length > 0) {
      this.image = randomElement(this.config.images);
    } else {
      this.image = undefined;
    }
  }
  updateConfig(config) {
    const previousConfig = this.config;
    this.config = {
      ...defaultConfig,
      ...config
    };
    this.config.changeFrequency = random(this.config.changeFrequency, this.config.changeFrequency * 1.5);

    // Update the radius if the config has changed, it won't gradually update on it's own
    if (this.params && !isEqual(this.config.radius, previousConfig?.radius)) {
      this.params.radius = random(...this.config.radius);
    }
    if (!isEqual(this.config.images, previousConfig?.images)) {
      this.selectImage();
    }
  }
  updateTargetParams() {
    this.params.nextSpeed = random(...this.config.speed);
    this.params.nextWind = random(...this.config.wind);
    if (this.image) {
      this.params.nextRotationSpeed = random(...this.config.rotationSpeed);
    }
  }
  update(canvas, framesPassed = 1) {
    const {
      x,
      y,
      rotation,
      rotationSpeed,
      nextRotationSpeed,
      wind,
      speed,
      nextWind,
      nextSpeed,
      radius
    } = this.params;

    // Update current location, wrapping around if going off the canvas
    this.params.x = (x + wind * framesPassed) % (canvas.offsetWidth + radius * 2);
    if (this.params.x > canvas.offsetWidth + radius) {
      this.params.x = -radius;
    }
    this.params.y = (y + speed * framesPassed) % (canvas.offsetHeight + radius * 2);
    if (this.params.y > canvas.offsetHeight + radius) {
      this.params.y = -radius;
    }

    // Apply rotation
    if (this.image) {
      this.params.rotation = (rotation + rotationSpeed) % 360;
    }

    // Update the wind, speed and rotation towards the desired values
    this.params.speed = lerp(speed, nextSpeed, 0.01);
    this.params.wind = lerp(wind, nextWind, 0.01);
    this.params.rotationSpeed = lerp(rotationSpeed, nextRotationSpeed, 0.01);
    if (this.framesSinceLastUpdate++ > this.config.changeFrequency) {
      this.updateTargetParams();
      this.framesSinceLastUpdate = 0;
    }
  }
  getImageOffscreenCanvas(image, size) {
    if (image instanceof HTMLImageElement && image.loading) return image;
    let sizes = Snowflake.offscreenCanvases.get(image);
    if (!sizes) {
      sizes = {};
      Snowflake.offscreenCanvases.set(image, sizes);
    }
    if (!(size in sizes)) {
      const canvas = document.createElement("canvas");
      canvas.width = size;
      canvas.height = size;
      canvas.getContext("2d")?.drawImage(image, 0, 0, size, size);
      sizes[size] = canvas;
    }
    return sizes[size] ?? image;
  }
  draw(ctx) {
    if (this.image) {
      // ctx.save()
      // ctx.translate(this.params.x, this.params.y)
      ctx.setTransform(1, 0, 0, 1, this.params.x, this.params.y);
      const radius = Math.ceil(this.params.radius);
      ctx.rotate(this.params.rotation * Math.PI / 180);
      ctx.drawImage(this.getImageOffscreenCanvas(this.image, radius), -Math.ceil(radius / 2), -Math.ceil(radius / 2), radius, radius);

      // ctx.restore()
    } else {
      ctx.beginPath();
      ctx.arc(this.params.x, this.params.y, this.params.radius, 0, 2 * Math.PI);
      ctx.fillStyle = this.config.color;
      ctx.closePath();
      ctx.fill();
    }
  }
}

/**
 * A utility function to create a collection of snowflakes
 * @param canvasRef A ref to the canvas element
 * @param amount The number of snowflakes
 * @param config The configuration for each snowflake
 */
const makeSnowflakesArray = (canvasRef, amount, config) => {
  if (!canvasRef) return [];
  const snowflakes = [];
  for (let i = 0; i < amount; i++) {
    snowflakes.push(new Snowflake(canvasRef, config));
  }
  return snowflakes;
};

/**
 * A utility hook to manage creating and updating a collection of snowflakes
 * @param canvasRef A ref to the canvas element
 * @param amount The number of snowflakes
 * @param config The configuration for each snowflake
 */
const createSnowFlakes = (canvasRef, amount, config) => {
  const [snowflakes, setSnowflakes] = solidJs.createSignal([]);

  // Handle change of amount
  solidJs.createEffect(solidJs.on(amount, () => {
    setSnowflakes(snowflakes => {
      const sizeDifference = amount() - snowflakes.length;
      if (sizeDifference > 0) {
        return [...snowflakes, ...makeSnowflakesArray(canvasRef(), sizeDifference, config)];
      }
      if (sizeDifference < 0) {
        return snowflakes.slice(0, amount());
      }
      return snowflakes;
    });
  }));

  // Handle change of config
  solidJs.createEffect(() => {
    deep.trackStore(config);
    setSnowflakes(snowflakes => snowflakes.map(snowflake => {
      snowflake.updateConfig(config);
      return snowflake;
    }));
  });
  return snowflakes;
};

/**
 * Returns the height and width of a HTML element, uses the `ResizeObserver` api if available to detect changes to the
 * size. Falls back to listening for resize events on the window.
 * @param ref A ref to the HTML element to be measured
 */
const createComponentSize = ref => {
  const [size, setSize] = solidJs.createSignal(getSize(ref()));
  const resizeHandler = () => {
    if (ref()) {
      setSize(getSize(ref()));
    }
  };
  solidJs.createEffect(() => {
    const {
      ResizeObserver
    } = window;
    if (!ref()) return;
    resizeHandler();
    if (typeof ResizeObserver === "function") {
      const resizeObserver = new ResizeObserver(resizeHandler);
      resizeObserver.observe(ref());
      solidJs.onCleanup(() => resizeObserver.disconnect());
    } else {
      window.addEventListener("resize", resizeHandler);
      solidJs.onCleanup(() => window.removeEventListener("resize", resizeHandler));
    }
  });
  return size;
};

const snowfallBaseStyle = {
  "pointer-events": "none",
  "background-color": "transparent",
  position: "absolute",
  top: 0,
  left: 0,
  width: "100%",
  height: "100%"
};

// Target of 60 frames per second
const targetFrameTime = 1000 / 60;

var _tmpl$ = /*#__PURE__*/web.template(`<canvas data-testid=SnowfallCanvas>`);
const Snowfall = props => {
  const [configProps, other] = solidJs.splitProps(props, ["color", "changeFrequency", "radius", "speed", "wind", "rotationSpeed", "images"], ["snowflakeCount", "style"]);
  const config = solidJs.mergeProps(defaultConfig, configProps);
  const mergedStyle = () => ({
    ...snowfallBaseStyle,
    ...(other.style || {})
  });
  const [canvasRef, setCanvasRef] = solidJs.createSignal(null);
  const ctx = solidJs.createMemo(() => canvasRef()?.getContext("2d"));
  const canvasSize = resizeObserver.createElementSize(canvasRef);
  let animationFrame = 0;
  let lastUpdate = Date.now();
  const snowflakes = createSnowFlakes(canvasRef, () => other.snowflakeCount, config);

  // no need for reactivity here
  const render = (framesPassed = 1) => {
    // Update the positions of the snowflakes
    snowflakes().forEach(snowflake => {
      snowflake.update(canvasRef(), framesPassed);
    });
    const context = ctx();

    // Render them if the canvas is available
    // for images
    context.setTransform(1, 0, 0, 1, 0, 0);
    context.clearRect(0, 0, canvasRef().offsetWidth, canvasRef().offsetHeight);
    snowflakes().forEach(snowflake => snowflake.draw(context));
  };
  const loop = () => {
    // Update based on time passed so that a slow frame rate won't slow down the snowflake
    const now = Date.now();
    const msPassed = Date.now() - lastUpdate;
    lastUpdate = now;

    // Frames that would have passed if running at 60 fps
    const framesPassed = msPassed / targetFrameTime;
    render(framesPassed);
    animationFrame = requestAnimationFrame(loop);
  };
  solidJs.onMount(() => {
    loop();
  });
  solidJs.onCleanup(() => cancelAnimationFrame(animationFrame));
  return (() => {
    var _el$ = _tmpl$();
    web.use(setCanvasRef, _el$);
    web.effect(_p$ => {
      var _v$ = canvasSize.height,
        _v$2 = canvasSize.width,
        _v$3 = mergedStyle();
      _v$ !== _p$.e && web.setAttribute(_el$, "height", _p$.e = _v$);
      _v$2 !== _p$.t && web.setAttribute(_el$, "width", _p$.t = _v$2);
      _p$.a = web.style(_el$, _v$3, _p$.a);
      return _p$;
    }, {
      e: undefined,
      t: undefined,
      a: undefined
    });
    return _el$;
  })();
};

exports.Snowfall = Snowfall;
exports.Snowflake = Snowflake;
exports.createComponentSize = createComponentSize;
exports.createSnowFlakes = createSnowFlakes;
exports.default = Snowfall;
exports.defaultConfig = defaultConfig;
exports.getSize = getSize;
exports.lerp = lerp;
exports.random = random;
exports.randomElement = randomElement;
//# sourceMappingURL=index.js.map
